@@ -65,6 +65,7 @@ TestGetRawDecomposition(void);
 
 static void TestAppendRestoreMiddle(void);
 static void TestGetEasyToUseInstance(void);
+static void TestAPICoverage(void);
 
 static const char* const canonTests[][3] = {
     /* Input*/                    /*Decomposed*/                /*Composed*/
@@ -156,9 +157,11 @@ void addNormTest(TestNode** root)
     addTest(root, &TestGetRawDecomposition, "tsnorm/cnormtst/TestGetRawDecomposition");
     addTest(root, &TestAppendRestoreMiddle, "tsnorm/cnormtst/TestAppendRestoreMiddle");
     addTest(root, &TestGetEasyToUseInstance, "tsnorm/cnormtst/TestGetEasyToUseInstance");
+    addTest(root, &TestAPICoverage, "tsnorm/cnormtst/TestAPICoverage");
 }
 
 static const char* const modeStrings[]={
+    "?",
     "UNORM_NONE",
     "UNORM_NFD",
     "UNORM_NFKD",
@@ -183,7 +186,7 @@ static void TestNormCases(UNormalizationMode mode,
         length2= unorm_normalize(source, -1, mode, 0, NULL, 0, &status2);
         if(neededLen!=length2) {
           log_err("ERROR in unorm_normalize(%s)[%d]: "
-                  "preflight length/NUL %d!=%d preflight length/srcLength\n",
+                  "preflight length/srcLength %d!=%d preflight length/NUL\n",
                   modeStrings[mode], (int)x, (int)neededLen, (int)length2);
         }
         if(status==U_BUFFER_OVERFLOW_ERROR)
@@ -192,14 +195,14 @@ static void TestNormCases(UNormalizationMode mode,
         }
         length2=unorm_normalize(source, u_strlen(source), mode, 0, result, UPRV_LENGTHOF(result), &status); 
         if(U_FAILURE(status) || neededLen!=length2) {
-            log_data_err("ERROR in unorm_normalize(%s/NUL) at %s:  %s - (Are you missing data?)\n",
+            log_data_err("ERROR in unorm_normalize(%s/srcLength) at %s:  %s - (Are you missing data?)\n",
                          modeStrings[mode], austrdup(source), myErrorName(status));
         } else {
             assertEqual(result, cases[x][expIndex], x);
         }
         length2=unorm_normalize(source, -1, mode, 0, result, UPRV_LENGTHOF(result), &status); 
         if(U_FAILURE(status) || neededLen!=length2) {
-            log_data_err("ERROR in unorm_normalize(%s/srcLength) at %s:  %s - (Are you missing data?)\n",
+            log_data_err("ERROR in unorm_normalize(%s/NUL) at %s:  %s - (Are you missing data?)\n",
                          modeStrings[mode], austrdup(source), myErrorName(status));
         } else {
             assertEqual(result, cases[x][expIndex], x);
@@ -483,6 +486,7 @@ static void TestQuickCheckStringResult()
                                                             UNORM_YES)
     {
       log_data_err("ERROR in NFD quick check for string at count %d - (Are you missing data?)\n", count);
+      free(d); free(c);
       return;
     }
 
@@ -490,6 +494,7 @@ static void TestQuickCheckStringResult()
                                                             UNORM_NO)
     {
       log_err("ERROR in NFC quick check for string at count %d\n", count);
+      free(d); free(c);
       return;
     }
 
@@ -505,6 +510,7 @@ static void TestQuickCheckStringResult()
                                                             UNORM_YES)
     {
       log_data_err("ERROR in NFKD quick check for string at count %d\n", count);
+      free(d); free(c);
       return;
     }
 
@@ -512,6 +518,7 @@ static void TestQuickCheckStringResult()
                                                             UNORM_YES)
     {
       log_err("ERROR in NFKC quick check for string at count %d\n", count);
+      free(d); free(c);
       return;
     }
 
@@ -641,7 +648,7 @@ void TestCheckFCD()
   {
     UBool fcdresult = unorm_quickCheck(datastr[count], 4, UNORM_FCD, &status);
     if (U_FAILURE(status)) {
-      log_data_err("unorm_quickCheck(FCD) failed: exception occured at data set %d - (Are you missing data?)\n", count);
+      log_data_err("unorm_quickCheck(FCD) failed: exception occurred at data set %d - (Are you missing data?)\n", count);
       break;
     }
     else {
@@ -660,7 +667,7 @@ void TestCheckFCD()
   for (count = 0; count < 50; count ++)
   {
     int size = 0;
-    UBool testresult = UNORM_YES;
+    UNormalizationCheckResult testresult = UNORM_YES;
     UChar data[20];
     UChar norm[100];
     UChar nfd[100];
@@ -673,7 +680,7 @@ void TestCheckFCD()
       normsize += unorm_normalize(data + size, 1, UNORM_NFD, 0,
                                   norm + normsize, 100 - normsize, &status);
       if (U_FAILURE(status)) {
-        log_data_err("unorm_quickCheck(FCD) failed: exception occured at data generation - (Are you missing data?)\n");
+        log_data_err("unorm_quickCheck(FCD) failed: exception occurred at data generation - (Are you missing data?)\n");
         break;
       }
       size ++;
@@ -683,7 +690,7 @@ void TestCheckFCD()
     nfdsize = unorm_normalize(data, size, UNORM_NFD, 0,
                               nfd, 100, &status);
     if (U_FAILURE(status)) {
-      log_data_err("unorm_quickCheck(FCD) failed: exception occured at normalized data generation - (Are you missing data?)\n");
+      log_data_err("unorm_quickCheck(FCD) failed: exception occurred at normalized data generation - (Are you missing data?)\n");
     }
 
     if (nfdsize != normsize || u_memcmp(nfd, norm, nfdsize) != 0) {
@@ -1408,6 +1415,11 @@ TestQuickCheckPerCP() {
         }
 
         length=unorm_normalize(s, length, UNORM_NFD, 0, nfd, UPRV_LENGTHOF(nfd), &errorCode);
+        if (U_FAILURE(errorCode)) {
+            log_data_err("%s:%d errorCode=%s\n", __FILE__, __LINE__, u_errorName(errorCode));
+            break;
+        }
+
         /* length-length == 0 is used to get around a compiler warning. */
         U16_GET(nfd, 0, length-length, length, lead);
         U16_GET(nfd, 0, length-1, length, trail);
@@ -1688,4 +1700,55 @@ TestGetEasyToUseInstance() {
     }
 }
 
+static void
+TestAPICoverage() {
+    UErrorCode errorCode = U_ZERO_ERROR;
+    const UNormalizer2 *n2 = unorm2_getNFDInstance(&errorCode);
+    if (U_FAILURE(errorCode)) {
+        log_err_status(errorCode, "unorm2_getNFDInstance() failed: %s\n", u_errorName(errorCode));
+        return;
+    }
+
+    if (!unorm2_hasBoundaryBefore(n2, u'C') || unorm2_hasBoundaryBefore(n2, 0x300)) {
+        log_err("unorm2_hasBoundaryBefore() returns unexpected results\n");
+    }
+
+    if (!unorm2_hasBoundaryAfter(n2, u'C') || unorm2_hasBoundaryAfter(n2, 0x300)) {
+        log_err("unorm2_hasBoundaryAfter() returns unexpected results\n");
+    }
+
+    if (!unorm2_isInert(n2, 0x50005) || unorm2_isInert(n2, 0x300)) {
+        log_err("unorm2_isInert() returns unexpected results\n");
+    }
+
+    errorCode = U_ZERO_ERROR;
+    if (!unorm2_isNormalized(n2, u"c\u0327\u0300", 3, &errorCode) ||
+            unorm2_isNormalized(n2, u"c\u0300\u0327", 3, &errorCode) ||
+            U_FAILURE(errorCode)) {
+        log_err("unorm2_isNormalized() returns unexpected results\n");
+    }
+
+    errorCode = U_ZERO_ERROR;
+    if (unorm2_quickCheck(n2, u"c\u0327\u0300", 3, &errorCode) == UNORM_NO ||
+            unorm2_quickCheck(n2, u"c\u0300\u0327", 3, &errorCode) == UNORM_YES ||
+            U_FAILURE(errorCode)) {
+        log_err("unorm2_quickCheck() returns unexpected results\n");
+    }
+
+    errorCode = U_ZERO_ERROR;
+    if (unorm2_spanQuickCheckYes(n2, u"c\u0327\u0300", 3, &errorCode) != 3 ||
+            unorm2_spanQuickCheckYes(n2, u"c\u0300\u0327", 3, &errorCode) != 1 ||
+            U_FAILURE(errorCode)) {
+        log_err("unorm2_spanQuickCheckYes() returns unexpected results\n");
+    }
+
+    errorCode = U_ZERO_ERROR;
+    UChar first[10] = { u'c', 0x300, 0, 0, 0, 0, 0, 0, 0, 0 };
+    int32_t length = unorm2_normalizeSecondAndAppend(
+        n2, first, 2, UPRV_LENGTHOF(first), u"\u0327d", 2, &errorCode);
+    if (U_FAILURE(errorCode) || length != 4 || u_strcmp(first, u"c\u0327\u0300d") != 0) {
+        log_err("unorm2_normalizeSecondAndAppend() returns unexpected results\n");
+    }
+}
+
 #endif /* #if !UCONFIG_NO_NORMALIZATION */
