@@ -21,6 +21,7 @@ NumberRangeFormatterTest::NumberRangeFormatterTest()
 
 NumberRangeFormatterTest::NumberRangeFormatterTest(UErrorCode& status)
         : USD(u"USD", status),
+          CHF(u"CHF", status),
           GBP(u"GBP", status),
           PTE(u"PTE", status) {
 
@@ -47,11 +48,15 @@ void NumberRangeFormatterTest::runIndexedTest(int32_t index, UBool exec, const c
         TESTCASE_AUTO(testCollapse);
         TESTCASE_AUTO(testIdentity);
         TESTCASE_AUTO(testDifferentFormatters);
+        TESTCASE_AUTO(testNaNInfinity);
         TESTCASE_AUTO(testPlurals);
         TESTCASE_AUTO(testFieldPositions);
         TESTCASE_AUTO(testCopyMove);
         TESTCASE_AUTO(toObject);
         TESTCASE_AUTO(testGetDecimalNumbers);
+        TESTCASE_AUTO(test21684_Performance);
+        TESTCASE_AUTO(test21358_SignPosition);
+        TESTCASE_AUTO(test21683_StateLeak);
     TESTCASE_AUTO_END;
 }
 
@@ -135,14 +140,14 @@ void NumberRangeFormatterTest::testBasic() {
             .numberFormatterBoth(NumberFormatter::with().unit(FAHRENHEIT).unitWidth(UNUM_UNIT_WIDTH_FULL_NAME)),
         Locale("fr-FR"),
         u"1–5\u00A0degrés Fahrenheit",
-        u"≈5\u00A0degrés Fahrenheit",
-        u"≈5\u00A0degrés Fahrenheit",
+        u"≃5\u00A0degrés Fahrenheit",
+        u"≃5\u00A0degrés Fahrenheit",
         u"0–3\u00A0degrés Fahrenheit",
-        u"≈0\u00A0degré Fahrenheit",
+        u"≃0\u00A0degré Fahrenheit",
         u"3–3\u202F000\u00A0degrés Fahrenheit",
         u"3\u202F000–5\u202F000\u00A0degrés Fahrenheit",
         u"4\u202F999–5\u202F001\u00A0degrés Fahrenheit",
-        u"≈5\u202F000\u00A0degrés Fahrenheit",
+        u"≃5\u202F000\u00A0degrés Fahrenheit",
         u"5\u202F000–5\u202F000\u202F000\u00A0degrés Fahrenheit");
 
     assertFormatRange(
@@ -150,14 +155,14 @@ void NumberRangeFormatterTest::testBasic() {
         NumberRangeFormatter::with(),
         Locale("ja"),
         u"1～5",
-        u"約 5",
-        u"約 5",
+        u"約5",
+        u"約5",
         u"0～3",
-        u"約 0",
+        u"約0",
         u"3～3,000",
         u"3,000～5,000",
         u"4,999～5,001",
-        u"約 5,000",
+        u"約5,000",
         u"5,000～5,000,000");
 
     assertFormatRange(
@@ -663,6 +668,29 @@ void NumberRangeFormatterTest::testDifferentFormatters() {
         u"5,000–5,000,000");
 }
 
+void NumberRangeFormatterTest::testNaNInfinity() {
+    IcuTestErrorCode status(*this, "testNaNInfinity");
+
+    auto lnf = NumberRangeFormatter::withLocale("en");
+    auto result1 = lnf.formatFormattableRange(-uprv_getInfinity(), 0, status);
+    auto result2 = lnf.formatFormattableRange(0, uprv_getInfinity(), status);
+    auto result3 = lnf.formatFormattableRange(-uprv_getInfinity(), uprv_getInfinity(), status);
+    auto result4 = lnf.formatFormattableRange(uprv_getNaN(), 0, status);
+    auto result5 = lnf.formatFormattableRange(0, uprv_getNaN(), status);
+    auto result6 = lnf.formatFormattableRange(uprv_getNaN(), uprv_getNaN(), status);
+    auto result7 = lnf.formatFormattableRange({"1000", status}, {"Infinity", status}, status);
+    auto result8 = lnf.formatFormattableRange({"-Infinity", status}, {"NaN", status}, status);
+
+    assertEquals("0 - inf", u"-∞ – 0", result1.toTempString(status));
+    assertEquals("-inf - 0", u"0–∞", result2.toTempString(status));
+    assertEquals("-inf - inf", u"-∞ – ∞", result3.toTempString(status));
+    assertEquals("NaN - 0", u"NaN–0", result4.toTempString(status));
+    assertEquals("0 - NaN", u"0–NaN", result5.toTempString(status));
+    assertEquals("NaN - NaN", u"~NaN", result6.toTempString(status));
+    assertEquals("1000 - inf", u"1,000–∞", result7.toTempString(status));
+    assertEquals("-inf - NaN", u"-∞ – NaN", result8.toTempString(status));
+}
+
 void NumberRangeFormatterTest::testPlurals() {
     IcuTestErrorCode status(*this, "testPlurals");
 
@@ -740,18 +768,19 @@ void NumberRangeFormatterTest::testFieldPositions() {
             3000,
             5000,
             expectedString);
-        static const UFieldPosition expectedFieldPositions[] = {
-            // field, begin index, end index
-            {UNUM_INTEGER_FIELD, 0, 1},
-            {UNUM_COMPACT_FIELD, 1, 2},
-            {UNUM_INTEGER_FIELD, 5, 6},
-            {UNUM_COMPACT_FIELD, 6, 7},
-            {UNUM_MEASURE_UNIT_FIELD, 8, 9}};
-        checkFormattedValue(
+        static const UFieldPositionWithCategory expectedFieldPositions[] = {
+            // category, field, begin index, end index
+            {UFIELD_CATEGORY_NUMBER_RANGE_SPAN, 0, 0, 2},
+            {UFIELD_CATEGORY_NUMBER, UNUM_INTEGER_FIELD, 0, 1},
+            {UFIELD_CATEGORY_NUMBER, UNUM_COMPACT_FIELD, 1, 2},
+            {UFIELD_CATEGORY_NUMBER_RANGE_SPAN, 1, 5, 7},
+            {UFIELD_CATEGORY_NUMBER, UNUM_INTEGER_FIELD, 5, 6},
+            {UFIELD_CATEGORY_NUMBER, UNUM_COMPACT_FIELD, 6, 7},
+            {UFIELD_CATEGORY_NUMBER, UNUM_MEASURE_UNIT_FIELD, 8, 9}};
+        checkMixedFormattedValue(
             message,
             result,
             expectedString,
-            UFIELD_CATEGORY_NUMBER,
             expectedFieldPositions,
             UPRV_LENGTHOF(expectedFieldPositions));
     }
@@ -765,19 +794,42 @@ void NumberRangeFormatterTest::testFieldPositions() {
             87654321,
             98765432,
             expectedString);
-        static const UFieldPosition expectedFieldPositions[] = {
-            // field, begin index, end index
-            {UNUM_GROUPING_SEPARATOR_FIELD, 2, 3},
-            {UNUM_GROUPING_SEPARATOR_FIELD, 6, 7},
-            {UNUM_INTEGER_FIELD, 0, 10},
-            {UNUM_GROUPING_SEPARATOR_FIELD, 13, 14},
-            {UNUM_GROUPING_SEPARATOR_FIELD, 17, 18},
-            {UNUM_INTEGER_FIELD, 11, 21}};
-        checkFormattedValue(
+        static const UFieldPositionWithCategory expectedFieldPositions[] = {
+            // category, field, begin index, end index
+            {UFIELD_CATEGORY_NUMBER_RANGE_SPAN, 0, 0, 10},
+            {UFIELD_CATEGORY_NUMBER, UNUM_GROUPING_SEPARATOR_FIELD, 2, 3},
+            {UFIELD_CATEGORY_NUMBER, UNUM_GROUPING_SEPARATOR_FIELD, 6, 7},
+            {UFIELD_CATEGORY_NUMBER, UNUM_INTEGER_FIELD, 0, 10},
+            {UFIELD_CATEGORY_NUMBER_RANGE_SPAN, 1, 11, 21},
+            {UFIELD_CATEGORY_NUMBER, UNUM_GROUPING_SEPARATOR_FIELD, 13, 14},
+            {UFIELD_CATEGORY_NUMBER, UNUM_GROUPING_SEPARATOR_FIELD, 17, 18},
+            {UFIELD_CATEGORY_NUMBER, UNUM_INTEGER_FIELD, 11, 21}};
+        checkMixedFormattedValue(
+            message,
+            result,
+            expectedString,
+            expectedFieldPositions,
+            UPRV_LENGTHOF(expectedFieldPositions));
+    }
+
+    {
+        const char16_t* message = u"Field position with approximately sign";
+        const char16_t* expectedString = u"~-100";
+        FormattedNumberRange result = assertFormattedRangeEquals(
+            message,
+            NumberRangeFormatter::withLocale("en-us"),
+            -100,
+            -100,
+            expectedString);
+        static const UFieldPositionWithCategory expectedFieldPositions[] = {
+            // category, field, begin index, end index
+            {UFIELD_CATEGORY_NUMBER, UNUM_APPROXIMATELY_SIGN_FIELD, 0, 1},
+            {UFIELD_CATEGORY_NUMBER, UNUM_SIGN_FIELD, 1, 2},
+            {UFIELD_CATEGORY_NUMBER, UNUM_INTEGER_FIELD, 2, 5}};
+        checkMixedFormattedValue(
             message,
             result,
             expectedString,
-            UFIELD_CATEGORY_NUMBER,
             expectedFieldPositions,
             UPRV_LENGTHOF(expectedFieldPositions));
     }
@@ -823,6 +875,9 @@ void NumberRangeFormatterTest::testCopyMove() {
     assertEquals("FormattedNumberRange move constructor", u"1,00–5,00 $US", result.toString(status));
     result = l1.formatFormattableRange(3, 6, status);
     assertEquals("FormattedNumberRange move assignment", u"3,00–6,00 $US", result.toString(status));
+    FormattedNumberRange fnrdefault;
+    fnrdefault.toString(status);
+    status.expectErrorAndReset(U_INVALID_STATE_ERROR);
 }
 
 void NumberRangeFormatterTest::toObject() {
@@ -903,6 +958,143 @@ void NumberRangeFormatterTest::testGetDecimalNumbers() {
     }
 }
 
+void NumberRangeFormatterTest::test21684_Performance() {
+    IcuTestErrorCode status(*this, "test21684_Performance");
+    LocalizedNumberRangeFormatter lnf = NumberRangeFormatter::withLocale("en");
+    // The following two lines of code should finish quickly.
+    lnf.formatFormattableRange({"-1e99999", status}, {"0", status}, status);
+    lnf.formatFormattableRange({"0", status}, {"1e99999", status}, status);
+}
+
+void NumberRangeFormatterTest::test21358_SignPosition() {
+    IcuTestErrorCode status(*this, "test21358_SignPosition");
+
+    // de-CH has currency pattern "¤ #,##0.00;¤-#,##0.00"
+    assertFormatRange(
+        u"Approximately sign position with spacing from pattern",
+        NumberRangeFormatter::with()
+            .numberFormatterBoth(NumberFormatter::with().unit(CHF)),
+        Locale("de-CH"),
+        u"CHF 1.00–5.00",
+        u"CHF≈5.00",
+        u"CHF≈5.00",
+        u"CHF 0.00–3.00",
+        u"CHF≈0.00",
+        u"CHF 3.00–3’000.00",
+        u"CHF 3’000.00–5’000.00",
+        u"CHF 4’999.00–5’001.00",
+        u"CHF≈5’000.00",
+        u"CHF 5’000.00–5’000’000.00");
+
+    // TODO(CLDR-13044): Move the sign to the inside of the number
+    assertFormatRange(
+        u"Approximately sign position with currency spacing",
+        NumberRangeFormatter::with()
+            .numberFormatterBoth(NumberFormatter::with().unit(CHF)),
+        Locale("en-US"),
+        u"CHF 1.00–5.00",
+        u"~CHF 5.00",
+        u"~CHF 5.00",
+        u"CHF 0.00–3.00",
+        u"~CHF 0.00",
+        u"CHF 3.00–3,000.00",
+        u"CHF 3,000.00–5,000.00",
+        u"CHF 4,999.00–5,001.00",
+        u"~CHF 5,000.00",
+        u"CHF 5,000.00–5,000,000.00");
+
+    {
+        LocalizedNumberRangeFormatter lnrf = NumberRangeFormatter::withLocale("de-CH");
+        UnicodeString actual = lnrf.formatFormattableRange(-2, 3, status).toString(status);
+        assertEquals("Negative to positive range", u"-2 – 3", actual);
+    }
+
+    {
+        LocalizedNumberRangeFormatter lnrf = NumberRangeFormatter::withLocale("de-CH")
+            .numberFormatterBoth(NumberFormatter::forSkeleton(u"%", status));
+        UnicodeString actual = lnrf.formatFormattableRange(-2, 3, status).toString(status);
+        assertEquals("Negative to positive percent", u"-2% – 3%", actual);
+    }
+
+    {
+        // TODO(CLDR-14111): Add spacing between range separator and sign
+        LocalizedNumberRangeFormatter lnrf = NumberRangeFormatter::withLocale("de-CH");
+        UnicodeString actual = lnrf.formatFormattableRange(2, -3, status).toString(status);
+        assertEquals("Positive to negative range", u"2–-3", actual);
+    }
+
+    {
+        LocalizedNumberRangeFormatter lnrf = NumberRangeFormatter::withLocale("de-CH")
+            .numberFormatterBoth(NumberFormatter::forSkeleton(u"%", status));
+        UnicodeString actual = lnrf.formatFormattableRange(2, -3, status).toString(status);
+        assertEquals("Positive to negative percent", u"2% – -3%", actual);
+    }
+}
+
+void NumberRangeFormatterTest::test21683_StateLeak() {
+    IcuTestErrorCode status(*this, "test21683_StateLeak");
+    UNumberRangeFormatter* nrf = nullptr;
+    UFormattedNumberRange* result = nullptr;
+    UConstrainedFieldPosition* fpos = nullptr;
+
+    struct Range {
+        double start;
+        double end;
+        const char16_t* expected;
+        int numFields;
+    } ranges[] = {
+        {1, 2, u"1\u20132", 4},
+        {1, 1, u"~1", 2},
+    };
+
+    UParseError* perror = nullptr;
+    nrf = unumrf_openForSkeletonWithCollapseAndIdentityFallback(
+        u"", -1,
+        UNUM_RANGE_COLLAPSE_AUTO,
+        UNUM_IDENTITY_FALLBACK_APPROXIMATELY,
+        "en", perror, status);
+    if (status.errIfFailureAndReset("unumrf_openForSkeletonWithCollapseAndIdentityFallback")) {
+        goto cleanup;
+    }
+
+    result = unumrf_openResult(status);
+    if (status.errIfFailureAndReset("unumrf_openResult")) { goto cleanup; }
+
+    for (auto range : ranges) {
+        unumrf_formatDoubleRange(nrf, range.start, range.end, result, status);
+        if (status.errIfFailureAndReset("unumrf_formatDoubleRange")) { goto cleanup; }
+
+        auto* formattedValue = unumrf_resultAsValue(result, status);
+        if (status.errIfFailureAndReset("unumrf_resultAsValue")) { goto cleanup; }
+
+        int32_t utf16Length;
+        const char16_t* utf16Str = ufmtval_getString(formattedValue, &utf16Length, status);
+        if (status.errIfFailureAndReset("ufmtval_getString")) { goto cleanup; }
+
+        assertEquals("Format", range.expected, utf16Str);
+
+        ucfpos_close(fpos);
+        fpos = ucfpos_open(status);
+        if (status.errIfFailureAndReset("ucfpos_open")) { goto cleanup; }
+
+        int numFields = 0;
+        while (true) {
+            bool hasMore = ufmtval_nextPosition(formattedValue, fpos, status);
+            if (status.errIfFailureAndReset("ufmtval_nextPosition")) { goto cleanup; }
+            if (!hasMore) {
+                break;
+            }
+            numFields++;
+        }
+        assertEquals("numFields", range.numFields, numFields);
+    }
+
+cleanup:
+    unumrf_close(nrf);
+    unumrf_closeResult(result);
+    ucfpos_close(fpos);
+}
+
 void  NumberRangeFormatterTest::assertFormatRange(
       const char16_t* message,
       const UnlocalizedNumberRangeFormatter& f,
